/****************************************************************************\
*
* MagnetiCE - Magnetic Scrolls Interpreter for Windows CE
*
* Windows CE Port/Front End written by Nick Lines (nick@chiark.com)
* Copyright (C) 2004 Nick Lines
*
* With thanks to Doug Boling for the wonderful 
* "Programming Microsoft Windows CE, 2nd edition" book
* from which elements of coding style have been inherited.
*
* Magnetic written by Niclas Karlsson <nkarlsso@abo.fi>,
*                     David Kinder <d.kinder@btinternet.com>,
*                     Stefan Meier <Stefan.Meier@if-legends.org> and
*                     Paul David Doherty <pdd@if-legends.org>
*
* Magnetic is Copyright (C) 1997-2003  Niclas Karlsson
*
*     This program is free software; you can redistribute it and/or modify
*     it under the terms of the GNU General Public License as published by
*     the Free Software Foundation; either version 2 of the License, or
*     (at your option) any later version.
*
*     This program is distributed in the hope that it will be useful,
*     but WITHOUT ANY WARRANTY; without even the implied warranty of
*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*     GNU General Public License for more details.
*
*     You should have received a copy of the GNU General Public License
*     along with this program; if not, write to the Free Software
*     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
*
*/


/*
*History
* 1.00 First release.  

* Todo:
* - add "more" status indication
* - resize
* - font requester
* - support for animation.  Given that I don't play animated games, this may not happen.	
*/

#include <windows.h>
#include <winbase.h>
#include <winuser.h>
#include <commctrl.h> //For command bar
#include <commdlg.h>  //Common dialogs
#include "..\generic\defs.h"
#include "resource.h"
#include "magnetiCE.h"

#define MAX_INPUT_SIZE	200
#define STATUS_SIZE	80
#define HISTORY_LINES	10
#define CHAR_BUFFER_SIZE		2000 //80 chars width x 25 
#define SCREEN_LINES	25  //15 at 8 point.

#define MSG_SHOW_GFX 1


//GLOBALS - Yes, I *know* they're wrong, but this is a quick and dirty hack
//to make things work.  I ought to encapsulate stuff in a structure at least,
//but you want the truth?  I couldn't be bothered ;-)

int loadFile(TCHAR *lpszFileName);
void lineOut(void);
HWND g_hWnd;
HWND g_hWndGfx;
HANDLE g_hGameThread;
HANDLE g_hBufferReadyForXmit;
HANDLE g_hEvtLoad;
HANDLE g_hBufferMutex;
HANDLE g_hStatusMutex;
HANDLE g_hGfxMutex;
HANDLE g_hMoreOver;
int	g_bCaret=0;
POINT g_caretPos;
int g_iInputPos;
HKEY g_myKey;


type8  g_anim;
type16 g_lGfxWidth;
type16 g_lGfxAlignedWidth;
type16 g_lGfxHeight;
type16 g_gfxPalette[16];
type8 *g_gfxData;
BITMAPINFO *g_gfxBMI;
type8 *g_gfxAlignedData;
HPALETTE g_hPal;
long	g_lGfxLeft;		 //Graphics Window position - left
long	g_lGfxTop;		 //Graphics Window position - top
double g_dGamma = 1.00;  //Gamma to be used in conversion

long g_lHeight;			//Height of client area in pixels
long g_lWidth;			//Width of client area in pixels
long g_lLines;			//Lines on screen
long g_lChars;       //Characters Per Line: this is approximate with proportional fonts!
long g_lFontHeight;		//Font height
long g_lFontWidth;		//Font Width (average character)
long g_lMoreFlag;
long g_lLinesSinceInput;
long g_lWantInput;		//Do we want input? 1 if last thing was ms_getchar, 0 else.
long g_lLoad;			//Load a game. 1 to quit and load...  (interlocked)
long g_lRunning;		//Is a game running?  1 if yes (interlocked)


HDC g_memoryDC;

//Game stuff
TCHAR szGfxName[MAX_PATH];
TCHAR szHintName[MAX_PATH];
TCHAR szFileName[MAX_PATH];



TCHAR *lpszHistoryBuffer;					//Actual buffer for data
TCHAR *lplpszHistoryLines[HISTORY_LINES];		//entries into buffer

TCHAR *lpszCharBuffer;
TCHAR **lplpszCharLines;

TCHAR *lpszLastLine;
TCHAR lpszInputLine[MAX_INPUT_SIZE];
TCHAR lpszStatus[STATUS_SIZE];

int g_lInputLines;
long g_lLinesPrinted;

const TCHAR szAppName[]=TEXT("MagnetiCE");
const TCHAR szAppGfxName[]=TEXT("MagGFX");


/*Registry Key Names*/
const TCHAR szGfxLeftKey[]=TEXT("GfxPositionLeft");
const TCHAR szGfxTopKey[]=TEXT("GfxPositionTop");
const TCHAR szGfxGammaKey[]=TEXT("GfxGamma");
const TCHAR szLast1[]=TEXT("RecentFile1");
const TCHAR szLast2[]=TEXT("RecentFile2");
const TCHAR szLast3[]=TEXT("RecentFile3");
const TCHAR szLast4[]=TEXT("RecentFile4");


HINSTANCE hInst;

int bGfxOn;

//Message dispatch table for our Main Window process
const struct decodeUINT mainMessages[] = {
	WM_CREATE, doCreateMain,
	WM_COMMAND, doCommandMain,
	WM_KEYUP, doKeysMain,
    WM_CHAR, doKeysMain,
	WM_INITMENUPOPUP,doInitMenuPopMain,
	WM_PAINT, doPaintMain,
	WM_SIZE,doSizeMain,
	WM_DESTROY, doDestroyMain,
	WM_SETFOCUS, doSetFocusMain,
	WM_KILLFOCUS, doKillFocusMain,
	WM_HELP, doHelpMain,
	WM_USER+MSG_SHOW_GFX,doShowGfx,
};

const struct decodeCMD mainCommandItems[]={
	IDM_FILE_OPEN, doMainCommandOpen,
	IDM_FILE_EXIT, doMainCommandExit,
	IDM_FILE_ABOUT, doMainCommandAbout,
	IDM_EDIT_UNDO, doMainCommandUndo,
	IDM_OPTIONS_FONT,doMainCommandFont,
};

//PROGRAM ENTRY POINT: WinMain
int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPWSTR lpCmdLine,int nShowCmd) {
	MSG msg;
	HACCEL hAccel;
	int iReturnCode=0;
	HWND hWndMain;

	if(iReturnCode=initApp(hInstance))  //InitApp registers our Window class
		return iReturnCode;

	hWndMain=initInstance(hInstance,lpCmdLine,nShowCmd); //Create main window etc
	if(hWndMain==0)
		return 0x10;

	hAccel=LoadAccelerators(hInst,MAKEINTRESOURCE(ID_ACCEL));
	g_hWnd=hWndMain; //Load our global (yukky)
//Message loop
	while(GetMessage(&msg,NULL,0,0)) {
		if(!TranslateAccelerator(hWndMain,hAccel,&msg))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

//Cleanup after getmessage returns zero value (end of loop for us)
	return terminateInstance(hInstance,msg.wParam);
}	


//APPLICATION INITIATION
int initApp(HINSTANCE hInstance) {
	WNDCLASS wc;
	
	wc.style=0;
	wc.lpfnWndProc=mainWndProc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hInstance=hInstance;
	wc.hIcon=NULL;
	wc.hCursor=LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground=(HBRUSH)GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName=NULL;
	wc.lpszClassName=szAppName;

	if(!(RegisterClass(&wc)))   // Return error if non-zero return
		return 1;

	wc.lpfnWndProc=gfxWndProc;
	wc.lpszClassName=szAppGfxName;
	if(!(RegisterClass(&wc)))   // Return error if non-zero return
		return 1;


	return 0;
}

//Initiate our particular application...
HWND initInstance(HINSTANCE hInstance, LPWSTR lpCmdLine, int nShowCmd) {
	HWND hWnd;
	HDC hDC;
	hInst=hInstance; //Store passed instance as global
	hWnd=CreateWindow(szAppName,TEXT("MagnetiCE1.0"),WS_VISIBLE,
					  CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,
					  NULL,NULL,hInstance,NULL);
	if(!IsWindow(hWnd))
		return 0;

	g_hBufferReadyForXmit=CreateEvent(NULL,FALSE,FALSE,NULL);
	g_hEvtLoad=CreateEvent(NULL,FALSE,FALSE,NULL);
	g_hMoreOver=CreateEvent(NULL,FALSE,FALSE,NULL);
	g_hBufferMutex=CreateMutex(NULL,FALSE,NULL);	
	g_hStatusMutex=CreateMutex(NULL,FALSE,NULL);	
	g_hGfxMutex=CreateMutex(NULL,FALSE,NULL);	

	g_hGameThread=CreateThread(NULL,0,gameThread,hWnd,0,NULL);
	if(!g_hGameThread)
	{
		DestroyWindow(hWnd);
		return 0;
	}

	getFontAndScreenMetrics(hWnd);  //Populate our globals
	if(!initiateBuffers(0)) {
		MessageBox(hWnd,TEXT("Could not allocate memory"),NULL,MB_OK);
		if (g_hGameThread) CloseHandle(g_hGameThread);
		DestroyWindow(hWnd);
		return 0;
	}

	g_lLinesSinceInput=g_lLinesPrinted=g_lInputLines=0;


	hDC=GetDC(hWnd);
	g_memoryDC=CreateCompatibleDC(hDC); //We use this for our font...
	ReleaseDC(hWnd,hDC);

	getSettingsFromRegistry();

	ShowWindow(hWnd,nShowCmd);
	UpdateWindow(hWnd);

	if(wcslen(lpCmdLine)>0) {
		wcscpy(&szFileName,lpCmdLine); //Hacky again :-) - global contains name of file to load.
		loadFile(lpCmdLine);			//I really have tried to make this unmaintainable!
	}

	return hWnd;
}

//Clean up after ourselves...
int terminateInstance(HINSTANCE hInstance, int nDefaultRC)
{	
	writeSettingsToRegistry();
	if(lpszHistoryBuffer)
		free(lpszHistoryBuffer);
	if(lpszCharBuffer)
		free(lpszCharBuffer);
	if(lpszLastLine)
		free(lpszLastLine);
	if(lplpszCharLines)
		free(lplpszCharLines);
	if(g_gfxBMI)
		free(g_gfxBMI);
	if(g_gfxAlignedData)
		free(g_gfxAlignedData);
	if(g_hGameThread)
		CloseHandle(g_hGameThread);
	if(g_memoryDC)
		DeleteDC(g_memoryDC);
	if(g_hMoreOver)
		CloseHandle(g_hBufferMutex);
	if(g_hBufferMutex)
		CloseHandle(g_hBufferMutex);
	if(g_hBufferReadyForXmit)
		CloseHandle(g_hBufferReadyForXmit);
	if(g_hEvtLoad)
		CloseHandle(g_hEvtLoad);
	if(g_hStatusMutex)
		CloseHandle(g_hStatusMutex);
	if(g_hGfxMutex)
		CloseHandle(g_hGfxMutex);
	
	return nDefaultRC;
}

//The Main Window Process
LRESULT CALLBACK mainWndProc(HWND hWnd,UINT wMsg, WPARAM wParam, LPARAM lParam) {
	INT i;
	for(i=0;i<dim(mainMessages); i++) {
		if (wMsg == mainMessages[i].code)
			return (*mainMessages[i].fxn)(hWnd,wMsg,wParam,lParam);
	}
	return DefWindowProc(hWnd,wMsg,wParam,lParam);
}

LRESULT doCreateMain(HWND hWnd,UINT wMsg, WPARAM wParam, LPARAM lParam) {
	HWND hWndCB;
	HICON hIcon;
	hWndCB = CommandBar_Create(hInst, hWnd, IDC_CMDBAR);
	CommandBar_InsertMenubar(hWndCB,hInst,IDR_MAINMENU,0);
	CommandBar_AddAdornments(hWndCB, CMDBAR_HELP,0);
	hIcon=(HICON)SendMessage(hWnd,WM_GETICON,0,0);
	if(hIcon==0) {
		hIcon=LoadImage(hInst,MAKEINTRESOURCE(IDI_ICON1),IMAGE_ICON,16,16,0);
		SendMessage(hWnd,WM_SETICON,FALSE,(LPARAM)hIcon);
	}
	return 0;
}

LRESULT doShowGfx(HWND hWnd,UINT wMsg, WPARAM wParam, LPARAM lParam) {
	RECT rect;	
	if(wParam) { //wParam is mode
		WaitForSingleObject(g_hGfxMutex,INFINITE);
		if(g_hWndGfx) {
			GetWindowRect(g_hWndGfx,&rect);
			g_lGfxLeft=rect.left;
			g_lGfxTop=rect.top;
			DestroyWindow(g_hWndGfx);
		}
		if(g_lGfxLeft>=(g_lWidth-50)) {
			g_lGfxLeft=g_lWidth-50;
		}
		if(g_lGfxLeft<0) {
			g_lGfxLeft=10;
		}
		if(g_lGfxTop>=(g_lHeight-50)) {
			g_lGfxTop=g_lHeight-50;
		}
		if(g_lGfxTop<0) {
			g_lGfxTop=10;
		}


		g_hWndGfx=CreateWindow(szAppGfxName,TEXT("MagneticGraphics"),WS_BORDER|WS_VISIBLE,g_lGfxLeft,g_lGfxTop,g_lGfxWidth, g_lGfxHeight,hWnd,NULL,hInst,NULL); //WS_SYSMENU|WS_CAPTION| for titlebar
		ReleaseMutex(g_hGfxMutex);
	}
	else 
		if (g_hWndGfx) {
			DestroyWindow(g_hWndGfx);
			g_hWndGfx=0;
		}
	return 0;
}


LRESULT doPaintMain(HWND hWnd,UINT wMsg, WPARAM wParam, LPARAM lParam) {
	RECT rect;
	PAINTSTRUCT ps;
	HDC hdc;
	int iEndLine;
	int iCurLine;
	int i;
	int j;
	int iFound;
	TCHAR *lpSt;
	TCHAR *lpScore;
	SIZE size;
	long lTextSize;
	long lInputSize;

	GetClientRect(hWnd,&rect);
	rect.top+=CommandBar_Height(GetDlgItem(hWnd,IDC_CMDBAR));
	rect.bottom=rect.top+g_lFontHeight;
	

	hdc=BeginPaint(hWnd,&ps);
//Take our mutex from getchar
	WaitForSingleObject(g_hBufferMutex,INFINITE);

	i=0;
	lpScore=0;
	iFound=0;
	lpSt=lplpszCharLines[0];
	j=wcslen(lplpszCharLines[0]);
	while((i<j) && (!iFound)) {
		if(lpszCharBuffer[i]==0x0009) {
			lpScore = lpszCharBuffer+i+1;
			iFound=1;
		}
		i++;
	}
	
	SetBkColor(hdc,0x00000000);
	SetTextColor(hdc,0x00FFFFFF);
	ExtTextOut(hdc,5,rect.top,ETO_OPAQUE,&rect,lplpszCharLines[0],i-1,NULL);
	if(iFound) {
		GetTextExtentPoint(hdc,lpScore,wcslen(lpScore),&size);
		ExtTextOut(hdc,rect.right-(size.cx+5),rect.top,0,NULL,lpScore,wcslen(lpScore),NULL); //
	}
	SetBkColor(hdc,0x00FFFFFF);
	SetTextColor(hdc,0x00000000);
	if(g_lLinesPrinted < (g_lLines-1)) //Not yet reached bottom of screen
		iEndLine=g_lLinesPrinted;
	else
		iEndLine=g_lLines;

	iCurLine=1;
	while(iCurLine < iEndLine)
	{
		//Print the line!
		ExtTextOut(hdc,5,rect.top+(iCurLine*g_lFontHeight),0,NULL,lplpszCharLines[((g_lLinesPrinted-(iEndLine-iCurLine)) % (g_lLines-1))+1],wcslen(lplpszCharLines[((g_lLinesPrinted-(iEndLine-iCurLine)) % (g_lLines-1)+1)])-1,NULL);
		iCurLine++;
	}
//release mutex
	ReleaseMutex(g_hBufferMutex);
//paint the current line if appropriate
	if((iCurLine>1) && (g_lWantInput)) {
		iCurLine--;
		GetTextExtentPoint(hdc,lplpszCharLines[((g_lLinesPrinted-(iEndLine-iCurLine)) % (g_lLines-1))+1],wcslen(lplpszCharLines[((g_lLinesPrinted-(iEndLine-iCurLine)) % (g_lLines-1)+1)])-1,&size);
		lTextSize=size.cx;
		rect.top+=(iCurLine*g_lFontHeight);
		rect.left=(lTextSize);
		rect.bottom=rect.top+g_lFontHeight;

		ExtTextOut(hdc,rect.left,rect.top,ETO_OPAQUE,&rect,lpszInputLine,g_iInputPos,NULL);
		GetTextExtentPoint(hdc,lpszInputLine,g_iInputPos,&size);
		lInputSize=size.cx;
		if(!g_bCaret) {
			ShowCaret(hWnd);
			g_bCaret=1;
		}
		SetCaretPos(rect.left+lInputSize,rect.top);		
	}
	if(!g_lWantInput)
		if(g_bCaret) {
			g_bCaret=0;
			HideCaret(hWnd);
		}

	
	EndPaint(hWnd,&ps);
	return 0;
}

LRESULT doKeysMain(HWND hWnd,UINT wMsg, WPARAM wParam, LPARAM lParam) {
//Keys are 25 left,26 up,27 right, 28 down
//21 pgup, 22 pgdn
//23 end 24 home
	static UINT lastMsg;
	static int iPosInHistory=0;
	char c;


	if (g_lMoreFlag) { //We get a WM_KEYUP after the WM_CHAR msg, so ignore the first...
		if (lastMsg!=WM_CHAR)	{
			InterlockedDecrement(&g_lMoreFlag);
			InterlockedExchange(&g_lLinesSinceInput,0);
			SetEvent(g_hMoreOver);
			return 0;
		}
		else {
			lastMsg=wMsg;
			return 0;
		}
	}
	lastMsg=wMsg; //Store this for above...

	if (wMsg==WM_KEYUP) {  //We want to know about keyups on cursor keys
		switch(wParam) {
		case 0x00000021: //page up
			break;
		case 0x00000022: //pgdn
			break;
		case 0x00000023: //end
			break;
		case 0x00000024: //home
			break;
		case 0x00000026: //up
			//Check to see if this is the first call
			if((iPosInHistory<g_lInputLines)&&(iPosInHistory<HISTORY_LINES))
			{
				iPosInHistory++;
				wcscpy(lpszInputLine,lplpszHistoryLines[(g_lInputLines-iPosInHistory) % HISTORY_LINES]);
				g_iInputPos=wcslen(lpszInputLine)-1;
				InvalidateRect(hWnd,NULL,0);
			}
			break;
		case 0x00000028: //Down
			switch(iPosInHistory) {
			case 0:
				break;
			case 1:
				lpszInputLine[0]=0x0000;
				g_iInputPos=0;
				iPosInHistory=0;
				InvalidateRect(hWnd,NULL,0);
				break;
			default:
				iPosInHistory--;
				wcscpy(lpszInputLine,lplpszHistoryLines[(g_lInputLines-iPosInHistory) % HISTORY_LINES]);
				g_iInputPos=wcslen(lpszInputLine)-1;
				InvalidateRect(hWnd,NULL,0);
				break;
			}
			break;
		}
	}

	if ((wMsg==WM_CHAR) && (g_lWantInput)){
		c=wParam;
		iPosInHistory=0;
		lpszInputLine[g_iInputPos++]=wParam;
		if(c==0x08) {//Backspace
			if (g_iInputPos>1)
				g_iInputPos-=2;
			else
				g_iInputPos=0;

		}
		if(c==0x09) //Tab...
		{
		}
		if(c==0x0D) //Return
		{
			g_iInputPos=0;
			WaitForSingleObject(g_hBufferMutex,INFINITE);
			wcscpy(&lplpszCharLines[((g_lLinesPrinted-1) % (g_lLines-1))+1][wcslen(lplpszCharLines[((g_lLinesPrinted-1) % (g_lLines-1))+1])-1],lpszInputLine);
//			lplpszCharLines[((g_lLinesPrinted-1) % (g_lLines-1))+1][wcslen(lplpszCharLines[((g_lLinesPrinted-1) % (g_lLines-1))+1])-1]=0x0000;
			ReleaseMutex(g_hBufferMutex);
			SetEvent(g_hBufferReadyForXmit);
		}
		InvalidateRect(hWnd,NULL,FALSE);
	}

	return 0;
}

LRESULT doSizeMain(HWND hWnd,UINT wMsg, WPARAM wParam, LPARAM lParam) {
	//TODO: implement sizing change (happens when user changes the size 
	//      due to changing Autohide of taskbar!)
	static int iTimesCalled=0;

	if(iTimesCalled) {
		MessageBox(hWnd,TEXT("The window has been resized.  This event is not handled correctly by Magnetic.\n\nPlease be aware that the game window is now the wrong size for the screen window!\n\nYou are strongly advised to restart :-)"),NULL,MB_OK);}
	iTimesCalled++;
	return 0;
}

LRESULT doSetFocusMain(HWND hWnd,UINT wMsg, WPARAM wParam, LPARAM lParam) {

	CreateCaret(hWnd,NULL,2,g_lFontHeight);
	SetCaretPos(10,150);
//	ShowCaret(hWnd);
	return 0;
}

LRESULT doHelpMain(HWND hWnd,UINT wMsg, WPARAM wParam, LPARAM lParam) {
	CreateProcess(TEXT("peghelp.exe"), TEXT("file:magnetice.htm#Main_Contents"), NULL, NULL, FALSE, 0, NULL, NULL, NULL, NULL);
	return 0;

}
LRESULT doKillFocusMain(HWND hWnd,UINT wMsg, WPARAM wParam, LPARAM lParam) {
	if(g_bCaret)
		HideCaret(hWnd);
	DestroyCaret();
	return 0;
}
LRESULT doInitMenuPopMain(HWND hWnd,UINT wMsg, WPARAM wParam, LPARAM lParam) {
	
	return 0;
}

LRESULT doCommandMain(HWND hWnd,UINT wMsg, WPARAM wParam, LPARAM lParam) {
	WORD idItem,wNotifyCode;
	HWND hWndCtl;
	int i;

	idItem=(WORD)LOWORD(wParam);
	wNotifyCode=(WORD)HIWORD(wParam);
	hWndCtl=(HWND)lParam;

	for(i=0;i<dim(mainCommandItems);i++) {
		if (idItem==mainCommandItems[i].code)
			return(*mainCommandItems[i].fxn)(hWnd,idItem,hWndCtl,wNotifyCode);
	}
	return 0;
}

LRESULT doDestroyMain(HWND hWnd,UINT wMsg, WPARAM wParam, LPARAM lParam) {
//Store position of Graphics window before destroying
//We can do this because we know that master is destroyed before child.
	
	RECT rect;
	if(g_hWndGfx) {
		GetWindowRect(g_hWndGfx,&rect);
		g_lGfxLeft=rect.left;
		g_lGfxTop=rect.top;
	}

	PostQuitMessage(0);
	return 0;
}

LPARAM doMainCommandOpen(HWND hWnd,WORD idItem,HWND hWndCtl, WORD wNotifyCode){

	const LPTSTR pszOpenFilter = TEXT("Magnetic Game Files (*.mag)\0*.mag\0All files (*.*)\0*.*\0\0");
	OPENFILENAME of;
	int rc;

	szFileName[0]='\0';
	memset(&of,0,sizeof(of));
	of.lStructSize=sizeof(of);
	of.hwndOwner=hWnd;
	of.lpstrFile=szFileName;
	of.nMaxFile=dim(szFileName);
	of.lpstrFilter=pszOpenFilter;
	of.Flags=0;
	if(rc=GetOpenFileName(&of))
		rc=loadFile(szFileName);
	else
		rc=MessageBox(hWnd,TEXT("open failed!"),TEXT("Oh"),MB_OK);
	return 0;
}

int loadFile(TCHAR *lpszFileName){
	TCHAR *lpszExt;

	wcscpy(szGfxName,lpszFileName);
	lpszExt=wcsrchr(szGfxName,'.');
	wcscpy(lpszExt,TEXT(".gfx"));

	wcscpy(szHintName,lpszFileName);
	lpszExt=wcsrchr(szHintName,'.');
	wcscpy(lpszExt,TEXT(".hnt"));
	
//if game running, signal ms_putchar to exit by kicking its signal
	InterlockedExchange(&g_lLoad,1);
	if(g_lRunning)
		SetEvent(g_hBufferReadyForXmit);
	SetEvent(g_hEvtLoad);
	return 1;
}


LPARAM doMainCommandAbout(HWND hWnd,WORD idItem,HWND hWndCtl, WORD wNotifyCode){
	DialogBox(hInst,TEXT("aboutDialog"),hWnd,aboutDialogProc);
	return 0;
}

BOOL CALLBACK aboutDialogProc(HWND hWnd,UINT wMsg,WPARAM wParam,LPARAM lParam) {
	switch(wMsg) {
		case WM_COMMAND:
			switch(LOWORD(wParam)) {
				case IDOK:
				case IDCANCEL:
					EndDialog(hWnd,0);
					return TRUE;
				default:
					break;
			}
			break;
		default:
			break;
	}

	return FALSE;
}

LPARAM doMainCommandExit(HWND hWnd,WORD idItem,HWND hWndCtl, WORD wNotifyCode){
	SendMessage(hWnd,WM_CLOSE,0,0);
	return 0;
}

LPARAM doMainCommandUndo(HWND hWnd,WORD idItem,HWND hWndCtl, WORD wNotifyCode){
	return 0;
}

LPARAM doMainCommandFont(HWND hWnd,WORD idItem,HWND hWndCtl, WORD wNotifyCode){
	MessageBox(hWnd,TEXT("Not implemented...  Drop me a mail if you want this done!"),NULL,MB_OK);
	return 0;
}



type8 ms_load_file(type8s * name, type8 * ptr, type16 size)
{
	FILE *fh;
	TCHAR *strLoadName;
	size_t i;
	strLoadName=(TCHAR *)malloc((strlen(name)+1)*sizeof(TCHAR));
	memset(strLoadName,0,sizeof(TCHAR)*(strlen(name)+1));
	for(i=0;i<strlen(name);i++)
		strLoadName[i]=name[i];
	if ((fh = _wfopen(strLoadName,TEXT("rb"))) == NULL) {
		free(strLoadName);
		return 1;
	}
	if (fread(ptr,1,size,fh) != size) {
		free(strLoadName);
		fclose(fh);
		return 1;
	}
	fclose(fh);
	free(strLoadName);
	return 0;

}

type8 ms_save_file(type8s * name, type8 * ptr, type16 size)
{
	FILE *fh;
	TCHAR *strSaveName;
	size_t i;

	strSaveName=(TCHAR *)malloc((strlen(name)+1)*sizeof(TCHAR));
	memset(strSaveName,0,sizeof(TCHAR)*(strlen(name)+1));
	for(i=0;i<strlen(name);i++)
		strSaveName[i]=name[i];
	if ((fh = _wfopen(strSaveName,TEXT("wb"))) == NULL){
		MessageBox(g_hWnd,TEXT("Open failed"),NULL,MB_OK);
		free(strSaveName);
		return 1;
	}
	if(fwrite(ptr,1,size,fh) != size) {
		MessageBox(g_hWnd,TEXT("Write failed"),NULL,MB_OK);
		free(strSaveName);
		fclose(fh);
		return 1;
	}
	fclose(fh);
	free(strSaveName);
	return 0;

}

void ms_statuschar(type8 c)
{
	static int iPos;
	switch(c) {
	case 0x0a:
		iPos=0;
		WaitForSingleObject(g_hBufferMutex,INFINITE);
		wcscpy(lplpszCharLines[0],lpszStatus);
		ReleaseMutex(g_hBufferMutex);
		memset(lpszStatus,0,STATUS_SIZE);
		break;
	case 0x09:
		lpszStatus[iPos++]=c;
		break;
	default:
		lpszStatus[iPos++]=c;
	}


	ReleaseMutex(g_hStatusMutex);		
}

type8 ms_showhints(struct ms_hint * hints)
{
	return 0;
}


void ms_putchar(type8 c)
{
	static int iBufPos=0;
	static int iLastSpace=0;
	static long lCursorPos=0;
	TCHAR *lpszTmp;

	TCHAR tcharTmp;
	SIZE size;
	InterlockedExchange(&g_lWantInput,0);

	tcharTmp=c; //Need to pass a TCHAR to gettextextentpoint...
	lpszLastLine[iBufPos++]=tcharTmp;
	if ((iBufPos>=g_lChars)||(c==0x0a))	{ 
		iBufPos=0;
		iLastSpace=0;
		lCursorPos=0;
		lineOut();
		return;
	}

	if (c==0x08) //Delete: used for inventory, and other things...
	{
		iBufPos-=2;
		lpszLastLine[iBufPos]=0x20;
	}
	GetTextExtentPoint(g_memoryDC,&tcharTmp,1,&size);
	lCursorPos+=size.cx;
	if (lCursorPos >= (g_lWidth - 10)) { //Back up to the last space
		lpszTmp=malloc(g_lChars * sizeof(TCHAR));
		wcscpy(lpszTmp,&lpszLastLine[iLastSpace]);
		lpszLastLine[iLastSpace]=0x00;
		lineOut();
		wcscpy(lpszLastLine,lpszTmp);
		free(lpszTmp);
		iBufPos=wcslen(lpszLastLine);
		iLastSpace=0;
		GetTextExtentPoint(g_memoryDC,lpszLastLine,iBufPos,&size);
		lCursorPos=size.cx;
	}

	if(c==0x20)
	{
		iLastSpace=iBufPos;
	}
}
void ms_flush(void)
{
	if(wcslen(lpszLastLine)>0)
		lineOut();
}

void lineOut(void)
{
	if(g_lMoreFlag)
		WaitForSingleObject(g_hMoreOver,INFINITE);


	//Take our mutex
	WaitForSingleObject(g_hBufferMutex,INFINITE);
	wcscpy(lplpszCharLines[(g_lLinesPrinted % (g_lLines-1))+1],lpszLastLine);
	g_lLinesPrinted++;
	ReleaseMutex(g_hBufferMutex);
	memset(lpszLastLine,0,sizeof(TCHAR) * g_lChars);
	InterlockedIncrement(&g_lLinesSinceInput);
	if(g_lLinesSinceInput >=(g_lLines-1))
		InterlockedIncrement(&g_lMoreFlag);
	InvalidateRect(g_hWnd,NULL,TRUE);
}

type8 ms_getchar(type8 trans)
{
	static int iBufPos=0;
	int nRC;
	char c=0;
	if(!g_lWantInput) {
		InterlockedExchange(&g_lWantInput,1);
		InvalidateRect(g_hWnd,NULL,FALSE);
	}


//Wait until we're signalled to go go go!
	if (iBufPos==0) {
		nRC=WaitForSingleObject(g_hBufferReadyForXmit,INFINITE);
		if(nRC!=WAIT_OBJECT_0){
			return 0;
		}
//is a load waiting?
		if(g_lLoad==1)
		{
			return 1; //Message to QUIT NOW!  Stop it!  Etc...
		}
	}



//Now we're here, we've got a line to return
//First, check for #undo.  Load c appropriately
	if(!wcsncmp(TEXT("#undo"),lpszInputLine,5))
		c=0;
	else
		c=(char)lpszInputLine[iBufPos++];
	
//Check to see if we're at EOL
	if ((c== 0x0D) || (!c)) {
		wcscpy(lplpszHistoryLines[g_lInputLines % HISTORY_LINES],lpszInputLine);
		memset(lpszInputLine,0,MAX_INPUT_SIZE*sizeof(TCHAR));
		g_lInputLines++;
		InterlockedExchange(&g_lLinesSinceInput,0);
		iBufPos = 0;
	}
	return (type8)c;
}

void ms_showpic(type32 c, type8 mode)
{
	//MessageBox(g_hWnd,TEXT("ShowPic"),NULL,MB_OK);

	if(mode) {
		WaitForSingleObject(g_hGfxMutex,INFINITE);
		g_gfxData=ms_extract(c,&g_lGfxWidth,&g_lGfxHeight,g_gfxPalette,&g_anim);
		renderPicture();
		ReleaseMutex(g_hGfxMutex);
	}
	SendMessage(g_hWnd,WM_USER+MSG_SHOW_GFX,mode,c);
}

void renderPicture(void)
{
	int i,r,g,b;
//	TCHAR strTmp[200];
//	TCHAR strShortTmp[8];
//	memset(strTmp,0,200*sizeof(TCHAR));
//	palette=malloc(sizeof(LOGPALETTE) + (16*(sizeof(PALETTEENTRY))));

	if(g_hPal)
		DeleteObject(g_hPal);

	if(!g_gfxBMI)
		g_gfxBMI=malloc(sizeof(BITMAPINFOHEADER) +  (16*(sizeof(RGBQUAD))));
	if(g_gfxAlignedData)
	{	
		free(g_gfxAlignedData);
		g_gfxAlignedData=0;
	}
	g_gfxBMI->bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	g_gfxBMI->bmiHeader.biWidth = g_lGfxWidth;
	g_gfxBMI->bmiHeader.biHeight = g_lGfxHeight * -1;	// top-down bitmap
	g_gfxBMI->bmiHeader.biPlanes = 1;
	g_gfxBMI->bmiHeader.biBitCount = 8;
	g_gfxBMI->bmiHeader.biCompression = BI_RGB;
	g_gfxBMI->bmiHeader.biSizeImage = 0;
	g_gfxBMI->bmiHeader.biXPelsPerMeter = 0;
	g_gfxBMI->bmiHeader.biYPelsPerMeter = 0;
	g_gfxBMI->bmiHeader.biClrUsed = 16;
	g_gfxBMI->bmiHeader.biClrImportant = 0;
	
	for(i=0;i<16;i++) {  //Entries are 3 bit.  To translate to 8 bit, multiply by 32.
		r=(g_gfxPalette[i] & 0x0700) >> 3;
		g=(g_gfxPalette[i] & 0x0070) << 1;
		b=(g_gfxPalette[i] & 0x0007) << 5;
		r=(int)pow(r,1/g_dGamma);
		r=(r>255 ? 255: r);
		g=(int)pow(g,1/g_dGamma);
		g=(g>255 ? 255: g);
		b=(int)pow(b,1/g_dGamma);
		b=(b>255 ? 255: b);

		g_gfxBMI->bmiColors[i].rgbRed=r;
		g_gfxBMI->bmiColors[i].rgbGreen=g;
		g_gfxBMI->bmiColors[i].rgbBlue=b;
		g_gfxBMI->bmiColors[i].rgbReserved=0;
//		swprintf(strShortTmp,TEXT("%.2x%.2x%.2x,"),r,g,b);
//		if(i==8)
//			wcscat(strTmp,TEXT("\n"));
//		wcscat(strTmp,strShortTmp);
//		palette->palPalEntry[i].peRed=r;
//		palette->palPalEntry[i].peBlue=b;
//		palette->palPalEntry[i].peGreen=g;
//		palette->palPalEntry[i].peFlags=0;
	}
//	MessageBox(g_hWnd,strTmp,NULL,MB_OK);
	
	g_lGfxAlignedWidth = g_lGfxWidth + ((g_lGfxWidth % 4)? 4-(g_lGfxWidth%4) : 0);
	g_gfxAlignedData=malloc(g_lGfxAlignedWidth*g_lGfxHeight);
//	palette->palVersion=0x300;
//	palette->palNumEntries=16;
//	g_hPal=CreatePalette(palette);
//	free(palette);
}

LRESULT CALLBACK gfxWndProc(HWND hWnd,UINT wMsg,WPARAM wParam,LPARAM lParam)
{
	HDC hDC;
	RECT rect;
	type8 *ppvBits;
	HBITMAP hBmp;
	HICON hIcon;
	HDC hDCmem;

	static POINT ptsBegin;
	static POINT ptsEnd;
	static long titleHeight;
	long x,y;
	static long lGfxWinHeight;
	static int bDragged=0;
	static int bWaitingForMove;
	PAINTSTRUCT paint;
	int i,j;
	
	switch(wMsg) {
		case WM_CREATE:
			GetClientRect(hWnd,&rect);
			lGfxWinHeight=g_lGfxHeight+(g_lGfxHeight-rect.bottom);
			titleHeight=g_lGfxHeight-rect.bottom;
			SetWindowPos(hWnd,HWND_TOP,0,0,g_lGfxWidth,lGfxWinHeight,SWP_NOMOVE);
			hIcon=(HICON)SendMessage(hWnd,WM_GETICON,0,0);
			if(hIcon==0) {
				hIcon=LoadImage(hInst,MAKEINTRESOURCE(IDI_ICON1),IMAGE_ICON,16,16,0);
				SendMessage(hWnd,WM_SETICON,FALSE,(LPARAM)hIcon);
			}
			return 0;
			break;
/*		case 0x84:  //NCHITTEST - WORKS on emulator, doesn't work on HPC!
			i=DefWindowProc(hWnd,wMsg,wParam,lParam);
			if(i==1)
				return 2; //
			else
				return i;
			break;
*/
		case WM_MOVE:
		//	g_lGfxLeft=LOWORD(lParam);
		//	g_lGfxTop=HIWORD(lParam);
			return 0;
		case WM_LBUTTONDOWN:
			bDragged=1;
			GetWindowRect(hWnd,&rect);
			SetCapture(hWnd);
			g_lGfxLeft=rect.left;
			g_lGfxTop=rect.top;
			ptsBegin.x=LOWORD(lParam);  
			ptsBegin.y=HIWORD(lParam);  
			return 0;
			break;

		case WM_MOUSEMOVE:
			if((wParam & MK_LBUTTON) && (bDragged)) {
				x=g_lGfxLeft+LOWORD(lParam)-ptsBegin.x;
				y=g_lGfxTop+HIWORD(lParam)-ptsBegin.y;
				if(x>=g_lWidth-10)
					x=g_lWidth-10;
				if(y>=g_lHeight-10)
					y=g_lHeight-10;
				g_lGfxLeft=x;
				g_lGfxTop=y;
				SetWindowPos(hWnd,HWND_TOP,g_lGfxLeft,g_lGfxTop,0,0,SWP_NOSIZE);					
			}
			return 0;
			break;
			
//		case WM_MOUSEMOVE: //Bit hacky this - we send a message of NCLBUTTONDOWN (non-client button down) with HTCAPTION as param
//			if((wParam & MK_LBUTTON) && (bDragged)){
//				ReleaseCapture();
//				SendMessage(hWnd,0xA1,2,0);  //0xa1=WM_NCLBUTTONDOWN, 2=HTCAPTION.
//			}
//			return 0;
//			break;
		case WM_LBUTTONUP:
			ReleaseCapture();
			bDragged=bWaitingForMove=0;
			return 0;
			break;
		case WM_COMMAND:
			switch(LOWORD(wParam)) {
				case IDOK:
				case IDCANCEL:
					DestroyWindow(hWnd);
					g_hWndGfx=0;
			}
			return 0;			break;
		case WM_PAINT:
			hDC=BeginPaint(hWnd,&paint);
			hDCmem=CreateCompatibleDC(hDC);
/* This section renders into a compatible bitmap... 
			hBmp=CreateCompatibleBitmap(hDC,g_lGfxAlignedWidth,g_lGfxHeight);
			SelectObject(hDCmem,hBmp);
			for (i = 0; i < g_lGfxHeight; i++)
			{
				for (j = 0; j < g_lGfxAlignedWidth; j++)
				{
					if (j < g_lGfxWidth) {
						cr[g_gfxData[(i*g_lGfxWidth)+j]]=SetPixel(hDCmem,j,i,(long)(g_gfxBMI->bmiColors[g_gfxData[(i*g_lGfxWidth)+j]].rgbRed + (g_gfxBMI->bmiColors[g_gfxData[(i*g_lGfxWidth)+j]].rgbGreen << 8) + (g_gfxBMI->bmiColors[g_gfxData[(i*g_lGfxWidth)+j]].rgbBlue << 16) ));		
					}
				}
			}
*/
/*Display the colours that we actually got from SetPixel()*/
/*
			for(i=0;i<8;i++) {
				swprintf(strTmpToo,TEXT("Wanted %8x,got %8x\n"),(long)(g_gfxBMI->bmiColors[i].rgbRed + (g_gfxBMI->bmiColors[i].rgbGreen << 8) + (g_gfxBMI->bmiColors[i].rgbBlue << 16)),cr[i]);
				wcscat(strTmp,strTmpToo);
			}
			MessageBox(g_hWnd,strTmp,NULL,MB_OK);
			memset(strTmp,0,sizeof(TCHAR) * 255);

			for(i=8;i<16;i++) {
				swprintf(strTmpToo,TEXT("Wanted %8x,got %8x\n"),(long)(g_gfxBMI->bmiColors[i].rgbRed + (g_gfxBMI->bmiColors[i].rgbGreen << 8) + (g_gfxBMI->bmiColors[i].rgbBlue << 16)),cr[i]);
				wcscat(strTmp,strTmpToo);
			}
			MessageBox(g_hWnd,strTmp,NULL,MB_OK);*/



/* This section uses DIBsection...*/

			hBmp=CreateDIBSection(hDCmem,g_gfxBMI,DIB_RGB_COLORS,&ppvBits,0,0);
			SelectObject(hDCmem,hBmp);
			for (i = 0; i < g_lGfxHeight; i++)
			{
				for (j = 0; j < g_lGfxAlignedWidth; j++)
				{
					if (j < g_lGfxWidth)
						ppvBits[(i*g_lGfxAlignedWidth)+j] = g_gfxData[(i*g_lGfxWidth)+j];
					else
						ppvBits[(i*g_lGfxAlignedWidth)+j] = 0;
				}
			}

			BitBlt(hDC,0,0,g_lGfxAlignedWidth,g_lGfxHeight,hDCmem,0,0,SRCCOPY);
			DeleteObject(hBmp);
			DeleteDC(hDCmem);
			EndPaint(hWnd,&paint);
			return 0;
			break;
		default:
			return DefWindowProc(hWnd,wMsg,wParam,lParam);
			break;
	}
}

void ms_fatal(type8s * txt)
{
	return;
}

DWORD WINAPI gameThread(LPVOID lpArg)
{
	HWND hWnd;
	int running;
	hWnd=(HWND)lpArg;

	while(1)	{
		if(running=ms_is_running()) {  //Deliberate =, not == !
		
			while(running) {
				if(g_lLoad) {
					ms_stop();
					ms_freemem();
					running=0;
					InterlockedExchange(&g_lRunning,0);
				}
				else {
					running=ms_rungame();
					InterlockedExchange(&g_lRunning,1);
				}
			}
		}
		else {
			//Wait on Load event
			WaitForSingleObject(g_hEvtLoad,INFINITE);
			ms_init((char *)szFileName,(char *)szGfxName,(char *)szHintName);
			InterlockedExchange(&g_lLoad,0);
		}
	}
	return 0;
}

void getFontAndScreenMetrics(HWND hWnd)
{
	RECT rect;
	TEXTMETRIC tm;
	HDC hDC;
	GetClientRect(hWnd,&rect);
	rect.top+=CommandBar_Height(GetDlgItem(hWnd,IDC_CMDBAR));
	
	hDC=GetDC(hWnd);
    if(GetTextMetrics(hDC,&tm)) {
		g_lFontHeight=tm.tmHeight;
		g_lFontWidth=tm.tmAveCharWidth;
		g_lWidth=rect.right-rect.left;
		g_lHeight=rect.bottom-rect.top;
		g_lLines=g_lHeight / g_lFontHeight; 
		g_lChars=(g_lWidth / g_lFontWidth) * 3; //Just in case we get iiiiiiiii ;-)
	}		
	ReleaseDC(hWnd,hDC);
}

int initiateBuffers(int iKeepData)
{
	int i;

    if(!iKeepData) //malloc...
	{
		
		if(!(lpszHistoryBuffer=(TCHAR *)malloc((MAX_INPUT_SIZE*HISTORY_LINES) * sizeof(TCHAR))))
			return 0;

		if(!(lpszLastLine=(TCHAR *)malloc(g_lChars * sizeof(TCHAR)))) {
			free(lpszHistoryBuffer);
			return 0;
		}


		if(!(lpszCharBuffer=(TCHAR *)malloc(g_lChars * g_lLines * sizeof(TCHAR)))) {
			free(lpszHistoryBuffer);
			free(lpszLastLine);
			return 0;
		}

		if(!(lplpszCharLines=(TCHAR **)malloc(g_lLines * sizeof(TCHAR *)))) {
			free(lpszHistoryBuffer);
			free(lpszLastLine);
			free(lpszCharBuffer);
			return 0;
		}
		
		memset(lpszHistoryBuffer,0,(MAX_INPUT_SIZE*HISTORY_LINES) * sizeof(TCHAR));
		memset(lpszCharBuffer,0,g_lChars * g_lLines * sizeof(TCHAR));
		memset(lplpszCharLines,0,g_lLines * sizeof(TCHAR *));

		for(i=0;i<g_lLines;i++)
			lplpszCharLines[i]=lpszCharBuffer+(i*g_lChars);

		for(i=0;i < HISTORY_LINES;i++)
			lplpszHistoryLines[i]=lpszHistoryBuffer+(i*MAX_INPUT_SIZE);
		return 1;
	}
	else
	{
		//Todo: implement reallocation of buffers
		return 0;
	}
}

void getSettingsFromRegistry(void) {
	long rc;
	long keyrc;
	DWORD size;
	DWORD type;
	rc=RegCreateKeyEx(HKEY_LOCAL_MACHINE,TEXT("software\\chiark\\magnetice"),0,NULL,0,0,NULL,&g_myKey,&keyrc);
	if(rc==ERROR_SUCCESS)
	{
		if(keyrc==REG_CREATED_NEW_KEY) {
		//Initialise our stuff to defaults...
			g_lGfxLeft=50;
			g_lGfxTop=30;
			g_dGamma=1.01;
		}
		else
		{
		//Read Reg keys
			size=sizeof(long);
			RegQueryValueEx(g_myKey,szGfxLeftKey,NULL,&type,(unsigned char *)&g_lGfxLeft,&size);
			size=sizeof(long);
			RegQueryValueEx(g_myKey,szGfxTopKey,NULL,&type,(unsigned char *)&g_lGfxTop,&size);
			size=sizeof(double);
			RegQueryValueEx(g_myKey,szGfxGammaKey,NULL,&type,(unsigned char *)&g_dGamma,&size);

		}
	}
}

void writeSettingsToRegistry(void) {
	//Write our registry settings
	RegSetValueEx(g_myKey,szGfxLeftKey,0,REG_DWORD,(unsigned char *)&g_lGfxLeft,sizeof(long));
	RegSetValueEx(g_myKey,szGfxTopKey,0,REG_DWORD,(unsigned char *)&g_lGfxTop,sizeof(long));
	RegSetValueEx(g_myKey,szGfxGammaKey,0,REG_BINARY,(unsigned char *)&g_dGamma,sizeof(double));
	RegCloseKey(g_myKey);
}